/* redir.c
 *
 * redirect-4-all, (C) Copyright 1999 by Crescent
 *
 */

#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/time.h>

/* defines */
#define OPT_PROXY_FTP 0x01
#define OPT_PROXY_IRC 0x02
#define OPT_BOUNCE    0x04
#define OPT_IDENT     0x08
#define OPT_CLEAR_CMD 0x10

#ifndef INADDR_NONE
#define INADDR_NONE -1
#endif

/* global variables */
struct sockaddr_in baddr_srv;

/* forward declarations */
int redirect(int,struct sockaddr_in*,int,int);

/*************************************************************************/

void strmcpy(char* dest, char* src, int len)
{ strncpy(dest,src,len);
  dest[len-1] = '\0';
}

int gets_sd(int sd, char* obuf, char* ibuf, int bufsize, int* bufused)
{ int   n;
  char* ptr;

  for (;;)
  { ptr = memchr(ibuf,'\n',*bufused);
    if (!ptr && (*bufused >= bufsize)) ptr = ibuf + bufsize - 1;
    if (ptr)
    { n = ptr - ibuf + 1;
      memmove(obuf,ibuf,n);
      obuf[n] = '\0';
      *bufused -= n;
      if (*bufused) memmove(ibuf,ibuf+n,*bufused);
      return 1;
    }
    n = read(sd,ibuf+(*bufused),bufsize-(*bufused));
    if (n <= 0)
    { if ((n < 0) && (errno == EAGAIN)) return 0;
      if (*bufused)
      { memmove(obuf,ibuf,*bufused);
        obuf[*bufused] = '\0';
        *bufused = 0;
        return 1;
      } else
      { if (!n) errno = EIO;
        return 0;
      }
    }
    *bufused += n;
  }
}

int gets_sd_to(int sd, char* obuf, char* ibuf, int bufsize, int* bufused,
  int timeout)
{ int i;

  i = gets_sd(sd,obuf,ibuf,bufsize,bufused);
  while (!i && (errno == EAGAIN))
  { if (timeout <= 0) break;
    timeout--;
    sleep(1);
    i = gets_sd(sd,obuf,ibuf,bufsize,bufused);
  }
  return i;
}

int fill_sockaddr(struct sockaddr_in* addr, char* caddr)
{ char            host[256];
  struct hostent* hp;
  char*           ptr;

  memset(addr,0,sizeof(struct sockaddr_in));
  addr->sin_family = AF_INET;
  ptr = strchr(caddr,':');
  if (ptr)
  { strmcpy(host,caddr,ptr-caddr < sizeof(host) ? ptr-caddr+1 : sizeof(host));
    addr->sin_port = htons(atoi(ptr+1));
    if (!strcmp(host,"*")) addr->sin_addr.s_addr = INADDR_ANY; else
    addr->sin_addr.s_addr = inet_addr(host);
    if (addr->sin_addr.s_addr == INADDR_NONE)
    { hp = gethostbyname(host);
      if (!hp) return -1;
      addr->sin_addr = ((struct in_addr*)hp->h_addr_list[0])[0];
    }
  } else
  { addr->sin_addr.s_addr = strchr(caddr,'.') ? inet_addr(caddr) : INADDR_NONE;
    if (addr->sin_addr.s_addr == INADDR_NONE)
    { addr->sin_addr.s_addr = INADDR_ANY;
      addr->sin_port = htons((unsigned short)atoi(caddr));
    } else addr->sin_port = 0;
  }
  return 0;
}

/* relays data from csock to dsock and vice versa */
int relay(int csock, int ssock)
{ char   buf[4096];
  fd_set fdset;
  int    i;
  int    maxfd;

  maxfd = csock > ssock ? csock+1 : ssock+1;
  FD_ZERO(&fdset);
  do
  { FD_SET(csock,&fdset);
    FD_SET(ssock,&fdset);
    i = select(maxfd,&fdset,NULL,NULL,NULL);
    if (i > 0)
    { if (FD_ISSET(csock,&fdset))
      { i = recv(csock,buf,sizeof(buf),0);
        if (i > 0) i = send(ssock,buf,i,0);
      }
      if ((i > 0) && FD_ISSET(ssock,&fdset))
      { i = recv(ssock,buf,sizeof(buf),0);
        if (i > 0) i = send(csock,buf,i,0);
      }
    }
  } while (i > 0);
  return 0;
}

/* relays ftp session from csock to ssock and intercepts PORT and PASV
   commands (= transparent ftp proxy) */
int relay_ftp(int csock, int ssock, FILE* cfileo, FILE* sfileo)
{ int                a1,a2,a3,a4;
  struct sockaddr_in addr;
  char               addrstr_c[32];
  char               addrstr_s[32];
  char               buf[2048];
  char               bufc[2048];
  int                bufc_used;
  char               bufs[2048];
  int                bufs_used;
  int                dsock_c;
  int                dsock_s;
  fd_set             fdset;
  int                i;
  long               l;
  int                maxfd;
  int                p1,p2;
  pid_t              pidc;
  pid_t              pids;
  char*              ptr;
  struct sockaddr_in rem_addr;

  maxfd = csock > ssock ? csock+1 : ssock+1;
  bufc_used = bufs_used = 0;
  pidc = pids = 0;

  /* get listen socket for client */
  i = sizeof(addr);
  if (getsockname(csock,(struct sockaddr*)&addr,&i) < 0) return -1;
  addr.sin_port = 0;
  dsock_c = socket(AF_INET,SOCK_STREAM,0);
  if (dsock_c < 0) return -1;
  if (bind(dsock_c,(struct sockaddr*)&addr,sizeof(addr)) < 0)
  { close(dsock_c);
    return -1;
  }
  if (listen(dsock_c,1) < 0)
  { close(dsock_c);
    return -1;
  }
  i = sizeof(struct sockaddr_in);
  if (getsockname(dsock_c,(struct sockaddr*)&addr,&i) < 0)
  { close(dsock_c);
    return -1;
  }
  l = ntohl(addr.sin_addr.s_addr);
  a1 = (l >> 24) & 0xFF;
  a2 = (l >> 16) & 0xFF;
  a3 = (l >> 8) & 0xFF;
  a4 = l & 0xFF;
  l = ntohs(addr.sin_port);
  p1 = (l >> 8) & 0xFF;
  p2 = l & 0xFF;
  sprintf(addrstr_c,"%i,%i,%i,%i,%i,%i",a1,a2,a3,a4,p1,p2);
  if (dsock_c >= maxfd) maxfd = dsock_c + 1;

  /* get listen socket for server */
  i = sizeof(addr);
  if (getsockname(ssock,(struct sockaddr*)&addr,&i) < 0) return -1;
  addr.sin_port = 0;
  dsock_s = socket(AF_INET,SOCK_STREAM,0);
  if (dsock_s < 0)
  { close(dsock_c);
    return -1;
  }
  if (bind(dsock_s,(struct sockaddr*)&addr,sizeof(addr)) < 0)
  { close(dsock_c);
    close(dsock_s);
    return -1;
  }
  if (listen(dsock_s,1) < 0)
  { close(dsock_c);
    close(dsock_s);
    return -1;
  }
  i = sizeof(struct sockaddr_in);
  if (getsockname(dsock_s,(struct sockaddr*)&addr,&i) < 0)
  { close(dsock_c);
    close(dsock_s);
    return -1;
  }
  l = ntohl(addr.sin_addr.s_addr);
  a1 = (l >> 24) & 0xFF;
  a2 = (l >> 16) & 0xFF;
  a3 = (l >> 8) & 0xFF;
  a4 = l & 0xFF;
  l = ntohs(addr.sin_port);
  p1 = (l >> 8) & 0xFF;
  p2 = l & 0xFF;
  sprintf(addrstr_s,"%i,%i,%i,%i,%i,%i",a1,a2,a3,a4,p1,p2);
  if (dsock_s >= maxfd) maxfd = dsock_s + 1;

  /* do relay */
  addr.sin_port = 0;
  FD_ZERO(&fdset);
  do
  { FD_SET(csock,&fdset);
    FD_SET(ssock,&fdset);
    FD_SET(dsock_c,&fdset);
    FD_SET(dsock_s,&fdset);
    i = select(maxfd,&fdset,NULL,NULL,NULL);
    if (i > 0)
    { /* data connection from client */
      if (FD_ISSET(dsock_c,&fdset))
      { i = sizeof(rem_addr);
        p1 = accept(dsock_c,(struct sockaddr*)&rem_addr,&i);
        pidc = addr.sin_port ? fork() : -1;
        if (pidc)
        { /* parent */
          close(p1);
        } else
        { /* child */
          close(csock);
          close(ssock);
          close(dsock_c);
          close(dsock_s);
          redirect(p1,&addr,0,1);
          close(p1);
          exit(0);
        }
      }

      /* data connection from server */
      if (FD_ISSET(dsock_s,&fdset))
      { i = sizeof(rem_addr);
        p1 = accept(dsock_s,(struct sockaddr*)&rem_addr,&i);
        pids = addr.sin_port ? fork() : -1;
        if (pids)
        { /* parent */
          close(p1);
        } else
        { /* child */
          close(csock);
          close(ssock);
          close(dsock_c);
          close(dsock_s);
          redirect(p1,&addr,0,0);
          close(p1);
          exit(0);
        }
      }

      /* command from client */
      if (FD_ISSET(csock,&fdset))
      { i = gets_sd(csock,buf,bufc,sizeof(bufc)-1,&bufc_used);
        if (i)
        { if (!strcmp(buf,"PASV\r\n") || !strcmp(buf,"PASV\n"))
          { /* intercept PASV command */
            i = fputs("PASV\r\n",sfileo) != EOF;
            if (i) i = gets_sd_to(ssock,buf,bufs,sizeof(bufs)-1,&bufs_used,30);
            if (i)
            { ptr = atoi(buf) == 227 ? strchr(buf,'(') : NULL;
              if (ptr)
              { if (sscanf(ptr,"(%i,%i,%i,%i,%i,%i)",&a1,&a2,&a3,&a4,&p1,&p2) == 6)
                { memset(&addr,0,sizeof(addr));
                  addr.sin_family = AF_INET;
                  l = (p1 << 8) + p2;
                  addr.sin_port = htons(l);
                  l = (a1 << 24) + (a2 << 16) + (a3 << 8) + a4;
                  addr.sin_addr.s_addr = htonl(l);
                  fprintf(cfileo,"227 Entering Passive Mode (%s).\r\n",
                    addrstr_c);
                } else ptr = NULL;
              }
              if (!ptr) i = fputs(buf,cfileo) != EOF;
            } else if (errno == EAGAIN) i = 1;
          } else if (!strncmp(buf,"PORT",4) && isspace(buf[4]))
          { /* intercept PORT command */
            if (sscanf(buf+5,"%i,%i,%i,%i,%i,%i",&a1,&a2,&a3,&a4,&p1,&p2) == 6)
            { memset(&addr,0,sizeof(addr));
              addr.sin_family = AF_INET;
              l = (p1 << 8) + p2;
              addr.sin_port = htons(l);
              l = (a1 << 24) + (a2 << 16) + (a3 << 8) + a4;
              addr.sin_addr.s_addr = htonl(l);
              fprintf(sfileo,"PORT %s\r\n",addrstr_s);
            } else i = fputs(buf,cfileo) != EOF;
          } else i = fputs(buf,sfileo) != EOF;
        }
      }

      /* reply from server */
      if (i && FD_ISSET(ssock,&fdset))
      { while (gets_sd(ssock,buf,bufs,sizeof(bufs)-1,&bufs_used))
          fputs(buf,cfileo);
        if (errno != EAGAIN) i = 0;
      }
    } else i = 0;
  } while (i);

  /* kill existing data connections */
  if (pidc) kill(pidc,SIGTERM);
  if (pids) kill(pids,SIGTERM);
  close(dsock_c);
  close(dsock_s);
  return 0;
}

/* looks up ident for csock and sends auth to ssock */
int bounce(int csock, int ssock, int options)
{ char               buf[1024];
  int                i;
  struct sockaddr_in loc_addr;
  int                loc_port;
  char*              ptr;
  struct sockaddr_in rem_addr;
  int                rem_port;
  int                sock;
  char               user[32];

  /* get local and remote port numbers */
  i = sizeof(loc_addr);
  if (getsockname(csock,(struct sockaddr*)&loc_addr,&i) < 0) return -1;
  i = sizeof(rem_addr);
  if (getpeername(csock,(struct sockaddr*)&rem_addr,&i) < 0) return -1;
  loc_port = ntohs(loc_addr.sin_port);
  rem_port = ntohs(rem_addr.sin_port);

  if (options & OPT_IDENT)
  { /* do ident lookup */

    /* connect to remote (client:113) */
    sock = socket(AF_INET,SOCK_STREAM,0);
    if (sock < 0) return -1;
    loc_addr.sin_port = 0;
    if (bind(sock,(struct sockaddr*)&loc_addr,sizeof(loc_addr)) < 0)
    { close(sock);
      return -1;
    }
    rem_addr.sin_port = htons(113);
    if (connect(sock,(struct sockaddr*)&rem_addr,sizeof(rem_addr)) < 0)
    { close(sock);
      return -1;
    }

    /* do query */
    sprintf(buf,"%i,%i\r\n",rem_port,loc_port);
    if (send(sock,buf,strlen(buf),0) < 0)
    { close(sock);
      return -1;
    }
    i = recv(sock,buf,sizeof(buf),0);
    close(sock);
    if (i <= 0) return -1;

    /* send authentication */
    loc_addr.sin_port = loc_port;
    rem_addr.sin_port = rem_port;
    if (sscanf(buf,"%i , %i : USERID :%*[^:]:%30s",&rem_port,&loc_port,user) == 3)
    { if ((loc_addr.sin_port == loc_port) && (rem_addr.sin_port == rem_port))
      { ptr = strchr(user,'\r');
        if (!ptr) ptr = strchr(user,'\n');
        if (ptr) *ptr = '\0';
        /* if remote addr is 127.0.0.1, send official ip address instead */
        if (rem_addr.sin_addr.s_addr == 0x0100007F)
        { i = sizeof(rem_addr);
          if (getsockname(ssock,(struct sockaddr*)&rem_addr,&i) < 0) return -1;
        }
        sprintf(buf,"AUTH %s@%s\r\n",user,inet_ntoa(rem_addr.sin_addr));
        send(ssock,buf,strlen(buf),0);
      }
    }
  } else
  { /* no ident lookup, send IP address only */

    /* if remote addr is 127.0.0.1, send official ip address instead */
    if (ntohl(rem_addr.sin_addr.s_addr) == 0x7F000001)
    { i = sizeof(rem_addr);
      if (getsockname(ssock,(struct sockaddr*)&rem_addr,&i) < 0) return -1;
    }
    sprintf(buf,"AUTH UNKNOWN@%s\r\n",inet_ntoa(rem_addr.sin_addr));
    send(ssock,buf,strlen(buf),0);
  }

  return 0;
}

/* structure for one IRC DCC proxy */
struct ircproxy_s
{ int            fd;      /* socket descriptor       */
  unsigned long  ipaddr;  /* originating ip address  */
  unsigned short port;    /* originating port number */
} ;

/* parses an irc message */
void irc_parse(char* s, char** p, int max, int* n, int ircmode)
{ char* q;

  q = s;
  while (*q && !isspace(*q)) q++;
  p[0] = s;
  if (*q) *(q++) = '\0';
  *n = 1;
  while (*q)
  { while (*q && isspace(*q)) q++;
    if (*q)
    { p[*n] = s = q;
      if (++(*n) >= max) break;
      if (ircmode && (*q == ':'))
      { while (*q) q++;
        if (*(q-1) == '\n') q--;
        if (*(q-1) == '\r') q--;
        *q = '\0';
        break;
      }
      while (*q && !isspace(*q)) q++;
      if (*q) *(q++) = '\0';
    }
  }
}

/* replaces DCC ip and port information */
void irc_dcc_replace(char* buf, char** p, int n, struct ircproxy_s** ircps,
  int* ircps_cnt, int* ircps_size, unsigned long ipaddr, int* maxfd, int cli)
{ struct sockaddr_in addr;
  int                fd;
  int                i;
  struct ircproxy_s* ptr;

  i = strlen(p[n]) - 1;
  if ((p[n][0] == ':') && (p[n][1] == '\x01') && (p[n][i] == '\x01'))
  { p[n][i] = '\0';
    irc_parse(p[n]+2,p+3,8,&n,0);
    if ((n >= 5) && !strcasecmp(p[3],"DCC") &&
       (((n == 5) && !strcasecmp(p[4],"CHAT")) ||
        ((n == 6) && !strcasecmp(p[4],"SEND"))))
    { if (*ircps_cnt >= *ircps_size)
      { i = *ircps_cnt + 8;
        ptr = (struct ircproxy_s*)realloc(*ircps,i*sizeof(struct ircproxy_s));
        if (ptr)
        { *ircps = ptr;
          *ircps_size = i;
        }
      } else ptr = *ircps;
      fd = ptr ? socket(AF_INET,SOCK_STREAM,0) : -1;
      if (fd >= 0)
      { addr.sin_family = AF_INET;
        addr.sin_addr.s_addr = htonl(ipaddr);
        addr.sin_port = 0;
        i = sizeof(addr);
        if (!bind(fd,(struct sockaddr*)&addr,i))
        { if (!getsockname(fd,(struct sockaddr*)&addr,&i))
          { listen(fd,1);
            if (fd >= *maxfd) *maxfd = fd + 1;
            (*ircps)[*ircps_cnt].fd = fd;
            sscanf(p[6],"%lu",&((*ircps)[*ircps_cnt].ipaddr));
            sscanf(p[7],"%hu",&((*ircps)[*ircps_cnt].port));
            (*ircps_cnt)++;
            if (cli)
            { if (n == 5)
              { sprintf(buf,"%s %s :\x01%s %s %s %lu %hu\x01\r\n",
                  p[0],p[1],p[3],p[4],p[5],ipaddr,ntohs(addr.sin_port));
              } else
              { sprintf(buf,"%s %s :\x01%s %s %s %lu %hu %s\x01\r\n",
                  p[0],p[1],p[3],p[4],p[5],ipaddr,ntohs(addr.sin_port),p[8]);
              }
            } else
            { if (n == 5)
              { sprintf(buf,"%s %s %s :\x01%s %s %s %lu %hu\x01\r\n",
                  p[0],p[1],p[2],p[3],p[4],p[5],ipaddr,ntohs(addr.sin_port));
              } else
              { sprintf(buf,"%s %s %s :\x01%s %s %s %lu %hu %s\x01\r\n",
                  p[0],p[1],p[2],p[3],p[4],p[5],ipaddr,ntohs(addr.sin_port),p[8]);
              }
            }
          } else close(fd);
        } else close(fd);
      }
    }
  }
}

/* relays irc session from csock to ssock and intercepts DCC CHAT and
   DCC SEND commands (= transparent irc proxy) */
int relay_irc(int csock, int ssock, FILE* cfileo, FILE* sfileo)
{ struct sockaddr_in addr;
  char               buf[2048];
  char               buf2[2048];
  char               bufc[2048];
  int                bufc_used;
  char               bufs[2048];
  int                bufs_used;
  int                fd;
  fd_set             fdset;
  int                i;
  int                n;
  unsigned long      ipaddr_c;
  unsigned long      ipaddr_s;
  struct ircproxy_s* ircps;
  int                ircps_cnt;
  int                ircps_size;
  int                maxfd;
  char*              p[16];
  pid_t              pid;

  maxfd = csock > ssock ? csock+1 : ssock+1;
  ircps = NULL;
  ircps_cnt = ircps_size = 0;
  bufc_used = bufs_used = 0;

  /* get ip addr for client */
  i = sizeof(addr);
  if (getsockname(csock,(struct sockaddr*)&addr,&i) < 0) return -1;
  ipaddr_c = ntohl(addr.sin_addr.s_addr);

  /* get ip addr for server */
  i = sizeof(addr);
  if (getsockname(ssock,(struct sockaddr*)&addr,&i) < 0) return -1;
  ipaddr_s = ntohl(addr.sin_addr.s_addr);

  /* do relay */
  do
  { FD_ZERO(&fdset);
    FD_SET(csock,&fdset);
    FD_SET(ssock,&fdset);
    for (i=0; i<ircps_cnt; i++) FD_SET(ircps[i].fd,&fdset);
    i = select(maxfd,&fdset,NULL,NULL,NULL);
    if (i > 0)
    { n = i;

      /* message from client */
      if (FD_ISSET(csock,&fdset))
      { n--;
        while (gets_sd(csock,buf,bufc,sizeof(bufc)-1,&bufc_used))
        { strcpy(buf2,buf);
          irc_parse(buf,p,8,&n,1);
          if ((n == 3) && !strcasecmp(p[0],"PRIVMSG")) irc_dcc_replace(buf2,
            p,2,&ircps,&ircps_cnt,&ircps_size,ipaddr_s,&maxfd,1);
          fputs(buf2,sfileo);
        }
        if (errno != EAGAIN) i = 0;
      }

      /* message from server */
      if (i && FD_ISSET(ssock,&fdset))
      { n--;
        while (gets_sd(ssock,buf,bufs,sizeof(bufs)-1,&bufs_used))
        { strcpy(buf2,buf);
          irc_parse(buf,p,8,&n,1);
          if ((n == 4) && !strcasecmp(p[1],"PRIVMSG")) irc_dcc_replace(buf2,
            p,3,&ircps,&ircps_cnt,&ircps_size,ipaddr_c,&maxfd,0);
          fputs(buf2,cfileo);
        }
        if (errno != EAGAIN) i = 0;
      }

      /* data connection */
      if (i && n) for (n=0; n<ircps_cnt; n++) if (FD_ISSET(ircps[n].fd,&fdset))
      { i = sizeof(addr);
        fd = accept(ircps[n].fd,(struct sockaddr*)&addr,&i);
        pid = fork();
        if (pid)
        { /* parent */
          close(fd);
          close(ircps[n].fd);
          if (n < --ircps_cnt) memmove(ircps+n,ircps+n+1,
            (ircps_cnt-n)*sizeof(struct ircproxy_s));
          n--;
        } else
        { /* child */
          addr.sin_family = AF_INET;
          addr.sin_addr.s_addr = htonl(ircps[n].ipaddr);
          addr.sin_port = htons(ircps[n].port);
          close(csock);
          close(ssock);
          for (i=0; i<ircps_cnt; i++) close(ircps[i].fd);
          redirect(fd,&addr,0,9);
          close(fd);
          exit(0);
        }
      }
    } else i = 0;
  } while (i);

  return 0;
}

/* redirects csock to rem_addr with FTP or IRC proxying and/or bouncing */
int redirect(int csock, struct sockaddr_in* rem_addr, int options, int bbind)
{ FILE*  cfileo;
  int    flags;
  FILE*  sfileo;
  int    ssock;

  ssock = socket(AF_INET,SOCK_STREAM,0);
  if (ssock < 0) return -1;
  if (bbind && baddr_srv.sin_addr.s_addr &&
      (bind(ssock,(struct sockaddr*)&baddr_srv,sizeof(baddr_srv)) < 0))
  { close(ssock);
    return -1;
  }
  if (connect(ssock,(struct sockaddr*)rem_addr,sizeof(struct sockaddr_in)) < 0)
  { close(ssock);
    return -1;
  }
  if (options & OPT_BOUNCE) bounce(csock,ssock,options);
  if (options & (OPT_PROXY_FTP | OPT_PROXY_IRC))
  { cfileo = fdopen(csock,"wt");
    sfileo = fdopen(ssock,"wt");
    if (cfileo && sfileo)
    { setvbuf(cfileo,NULL,_IOLBF,0);
      setvbuf(sfileo,NULL,_IOLBF,0);
      flags = fcntl(ssock,F_GETFL);
      if (flags >= 0) fcntl(ssock,F_SETFL,flags | O_NONBLOCK);
      if (options & OPT_PROXY_FTP)
      { if (relay_ftp(csock,ssock,cfileo,sfileo) < 0) relay(csock,ssock);
      } else if (options & OPT_PROXY_IRC)
      { flags = fcntl(csock,F_GETFL);
        if (flags >= 0) fcntl(csock,F_SETFL,flags | O_NONBLOCK);
        if (relay_irc(csock,ssock,cfileo,sfileo) < 0) relay(csock,ssock);
      } else relay(csock,ssock);
    } else relay(csock,ssock);
  } else relay(csock,ssock);
  close(ssock);
  return 0;
}

int usage(char* argv0)
{ printf(
    "usage: %s [-b <n>] [-s <l_addr>] [-cfhiv] [<l_addr>:]<l_port> [<r_addr>:]<r_port>\n"
    "       -b = bounce to ftp4all, n=do ident lookup (not with -i)\n"
    "       -c = clear command line\n"
    "       -f = proxy ftp data connections (not with -i)\n"
    "       -h = print this message\n"
    "       -i = proxy irc dcc connections (not with -b or -f)\n"
    "       -s = bind to l_addr when connecting to r_addr\n"
    "       -v = print version\n"
    ,argv0);
  return -1;
}

int main(int argc, char** argv)
{ struct sockaddr_in cli_addr;
  int                csock;
  int                i;
  struct sockaddr_in loc_addr;
  int                options = 0;
  pid_t              pid;
  struct sockaddr_in rem_addr;
  int                ssock;

  /* print startup message */
  printf(
    "REDIRECT4ALL 1.3 (C) Copyright 1999 by Crescent (crescent@gmx.net)\n"
    "  TCP port redirector with FTP and IRC proxying and FTP ident bouncing\n"
    "  This program is free software and distributed under GNU PUBLIC LICENSE\n");

  /* evaluate command line arguments */
  fill_sockaddr(&baddr_srv,"0");
  while ((i = getopt(argc,argv,"b:cfhis:v")) > 0) switch (i)
  { case 'b' :
      options |= OPT_BOUNCE;
      if (atoi(optarg)) options |= OPT_IDENT;
      break;
    case 'c' :
      options |= OPT_CLEAR_CMD;
      break;
    case 'f' :
      options |= OPT_PROXY_FTP;
      break;
    case 'h' :
      usage(argv[0]);
      return 0;
    case 'i' :
      options |= OPT_PROXY_IRC;
      break;
    case 's' :
      if (fill_sockaddr(&baddr_srv,optarg) < 0)
      { fprintf(stderr,"bad address: %s\n",optarg);
        return -1;
      }
      break;
    case 'v' :
      printf("%s version 1.3 (16/Oct/1999)\n",argv[0]);
      return 0;
    default :
      return usage(argv[0]);
  }
  if ((optind+2 != argc) ||
      ((options & OPT_PROXY_FTP) && (options & OPT_PROXY_IRC)) ||
      ((options & OPT_BOUNCE) && (options & OPT_PROXY_IRC)))
    return usage(argv[0]);
  if (fill_sockaddr(&loc_addr,argv[optind]) < 0)
  { fprintf(stderr,"bad address: %s\n",argv[optind]);
    return -1;
  }
  if (fill_sockaddr(&rem_addr,argv[optind+1]) < 0)
  { fprintf(stderr,"bad address: %s\n",argv[optind+1]);
    return -1;
  }

  /* allocate local socket */
  close(0);
  ssock = socket(AF_INET,SOCK_STREAM,0);
  if (ssock < 0)
  { fprintf(stderr,"unable to get socket: %s\n",strerror(errno));
    return -1;
  }
  if (bind(ssock,(struct sockaddr*)&loc_addr,sizeof(loc_addr)) < 0)
  { fprintf(stderr,"unable to bind to %s: %s\n",argv[optind],strerror(errno));
    return -1;
  }
  if (listen(ssock,16) < 0)
  { fprintf(stderr,"unable to listen on socket: %s\n",strerror(errno));
    return -1;
  }

  /* clear command line */
  if (options & OPT_CLEAR_CMD) for (i=1; i<argc; i++)
    memset(argv[i],0,strlen(argv[i]));

  /* go daemon */
  pid = fork();
  if (pid < 0)
  { fprintf(stderr,"unable to fork: %s\n",strerror(errno));
    return -1;
  } else if (pid)
  { printf("%s successfully set up and ready to redirect connections\n",
      argv[0]);
    return 0;
  }
  close(1);
  close(2);
  setsid();
  signal(SIGCLD,SIG_IGN);

  /* accept connections */
  for (;;)
  { i = sizeof(cli_addr);
    csock = accept(ssock,(struct sockaddr*)&cli_addr,&i);
    if (csock >= 0)
    { i = fork();
      if (i)
      { /* parent */
        close(csock);
      } else
      { /* child */
        close(ssock);
        redirect(csock,&rem_addr,options,1);
        close(csock);
        return 0;
      }
    }
  }

  /* notreached */
  return 0;
}

/* EOF */
